#!/usr/bin/env python3.11
import sys
import socket
import os
import boto3


def get_public_keys_local(connecting_user: str, ssh_home_dir: str) -> str:
    host_name = socket.getfqdn()
    key_dir = f"{ssh_home_dir}/{host_name}"
    key_path = f"{key_dir}/{connecting_user}-cert.pub"
    if not os.path.isdir(key_dir):
        os.makedirs(key_dir)
    if not os.path.exists(key_path):
        return ""
    with open(key_path, mode="rt", encoding="utf-8") as key_file:
        return key_file.read()


def get_public_keys_s3(connecting_user: str, bucket_name: str) -> str:
    host_name = socket.getfqdn()
    s3_client = boto3.client("s3")
    response = s3_client.get_object(
        Bucket=bucket_name, Key=f"{host_name}/{connecting_user}-cert.pub"
    )
    ssh_key = response["Body"].read().decode()
    return ssh_key


if __name__ == "__main__":
    CONNECTING_USER = sys.argv[1]
    ORIGIN = sys.argv[2]
    if ORIGIN == "local":
        print(get_public_keys_local(CONNECTING_USER, sys.argv[3]))
    elif ORIGIN == "s3":
        print(get_public_keys_s3(CONNECTING_USER, sys.argv[3]))



"Data stores that specifies where a Server stores its keys"
import os
from abc import ABC, abstractmethod
from tempfile import TemporaryDirectory
import boto3
from key_switcheroo.util import get_user_path


class DataStore(ABC):
    "A server uses a DataStore to get public keys from somewhere."

    @abstractmethod
    def get_sshd_config_line(self):
        "Return the config line that the server will add to the sshd_config"

    @abstractmethod
    def delete_key(self, host: str, user: str):
        "Delete the key for the host/user"

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_t, exc_v, exc_tb):
        pass


class S3DataStore(DataStore):
    "Store the public keys in an S3 bucket"

    def __init__(self, _s3_bucket_name: str, temp: bool = False):
        self._s3_bucket_name = _s3_bucket_name
        self.temp = temp

    @property
    def s3_bucket_name(self):
        "The name of the bucket in which the keys are stored"
        return self._s3_bucket_name

    def get_sshd_config_line(self) -> str:
        return f"s3 {self.s3_bucket_name}"

    def delete_key(self, host: str, user: str):
        s3_client = boto3.client("s3")
        s3_client.delete_object(Bucket=self.s3_bucket_name, Key=f"{host}/{user}")

    def __enter__(self):
        pass

    def __exit__(self, exc_t, exc_v, exc_tb):
        if self.temp:
            s3_client = boto3.client("s3")
            objects = s3_client.list_objects_v2(Bucket=self.s3_bucket_name)["Contents"]
            delete_objects = [{"Key": bucket_obj["Key"]} for bucket_obj in objects]  # type: ignore
            s3_client.delete_objects(
                Bucket=self.s3_bucket_name,
                Delete={"Objects": delete_objects},  # type: ignore
            )


class FileSystemDataStore(DataStore):
    "Stores keys in a file system"

    def __init__(self, temp: bool = False):
        self.temp = temp
        home_dir = f"{get_user_path()}/.ssh"
        if temp:
            self._dir: str | TemporaryDirectory = (
                TemporaryDirectory(  # pylint: disable=consider-using-with
                    prefix="ssh-keys-", dir=home_dir
                )
            )
        else:
            self._dir = home_dir

    @property
    def dir(self) -> str:
        if isinstance(self._dir, str):
            return self._dir
        return self._dir.name

    def get_sshd_config_line(self) -> str:
        return f"local {self.dir}"

    def delete_key(self, host: str, user: str):
        os.remove(f"{self.dir}/{host}/{user}")

    def __enter__(self):
        if isinstance(self._dir, str):
            if not os.path.isdir(self.dir):
                os.mkdir(self.dir)
        else:
            self._dir.__enter__()

    def __exit__(self, exc_t, exc_v, exc_tb):
        if isinstance(self._dir, TemporaryDirectory):
            self._dir.__exit__(None, None, None)



from abc import ABC, abstractmethod
import os
import shutil
import boto3
from key_switcheroo.custom_keygen import (
    generate_private_public_key_in_file,
    generate_private_public_key,
)
from key_switcheroo.util import get_user_path, get_username


def _ensure_ssh_home_exists():
    ssh_home = f"{get_user_path()}/.ssh"
    if not os.path.isdir(ssh_home):
        os.makedirs(ssh_home)


class Publisher(ABC):
    """Abstract key publisher base class"""

    @abstractmethod
    def publish_new_key(self) -> str:
        """Abstract method for publishing a new public key"""


class S3Publisher(Publisher):
    """S3 Publisher class"""

    def __init__(self, bucket_name: str, host: str, user_id: str):
        self.bucket_name = bucket_name
        self.host = host
        self.user_id = user_id

    def publish_new_key(self) -> str:
        # Generate new public/private key pair
        private_key, public_key = generate_private_public_key()
        _ensure_ssh_home_exists()
        # Store the new public key in S3 bucket
        s3_client = boto3.client("s3")
        s3_client.put_object(
            Body=public_key,
            Bucket=self.bucket_name,
            Key=f"{self.host}/{self.user_id}-cert.pub",
        )

        # Store the private key on the local machine
        private_key_dir = f"{get_user_path()}/.ssh/{self.host}"
        if not os.path.isdir(private_key_dir):
            os.makedirs(private_key_dir)
        private_key_path = f"{private_key_dir}/{self.user_id}"
        with open(private_key_path, "wb") as private_out:
            private_out.write(private_key)
        shutil.chown(private_key_path, user=get_username(), group=-1)
        os.chmod(private_key_path, 0o600)

        return public_key.decode()


class LocalPublisher(Publisher):
    """Local Publisher class"""

    def __init__(self, host: str, user_id: str):
        self.host = host
        self.user_id = user_id

    def publish_new_key(self) -> str:
        user_path = os.path.expanduser("~")
        _ensure_ssh_home_exists()
        _, public_key = generate_private_public_key_in_file(
            f"{user_path}/.ssh/{self.host}",
            private_key_name=self.user_id,
            public_key_name=f"{self.user_id}-cert.pub",
        )
        return public_key.decode()



"""Utility functions"""
import os
from socket import getservbyport
from random import randint
from psutil import Process


def get_open_port() -> int:
    "Returns a random open port, starting at 1024"
    start_port = 1024
    all_primary_ports = list(range(start_port, start_port + 100))
    last_selectable_port = all_primary_ports[len(all_primary_ports) - 1]

    def select_new_port() -> int:
        nonlocal all_primary_ports
        if len(all_primary_ports) == 0:  # Out of ports
            nonlocal last_selectable_port
            # Create a new port range to choose from
            all_primary_ports = list(
                range(last_selectable_port + 1, last_selectable_port + 1001)
            )
            last_selectable_port += 100
        # Choose a new port
        new_port_index = randint(0, len(all_primary_ports) - 1)
        # Remove it from our options, so we dont choose it again
        del all_primary_ports[new_port_index]
        return all_primary_ports[new_port_index]

    # Select a new port until we find an open one
    while True:
        selected_port = select_new_port()
        try:
            getservbyport(selected_port)
        except OSError:
            return selected_port


def get_default_authorized_keys_path() -> str:
    "Returns the default authorized keys path - ~/.ssh/authorized_keys"
    return f"{get_username()}/.ssh/authorized_keys"


def get_user_path() -> str:
    "Returns the user"
    return os.path.expanduser("~")


def get_username() -> str:
    "Returns the username on linux system (assuming running from somewhere inside the user path)"
    user_path = get_user_path()
    user_path_components = user_path.split("/")
    return user_path_components[len(user_path_components) - 1]


def get_process_running_with_pid(pid: int) -> str:
    "Returns the name of the process running with the given process id"
    process = Process(pid)
    return process.name()



from argparse import ArgumentParser
from key_switcheroo.publisher.key_publisher import LocalPublisher, S3Publisher


def create_argument_parser() -> ArgumentParser:
    argument_parser = ArgumentParser(
        prog="key_publisher",
        description="Creates public/private SSH keys and publishes "
        + "the public key either locally or to S3 (default is S3)",
        epilog="Thanks for using key_publisher! :)",
    )

    argument_parser.add_argument("hostname")
    argument_parser.add_argument("user")

    argument_parser.add_argument(
        "-ds",
        "--datastore",
        choices=["s3", "local"],
        default="s3",
        help="choose where to store the public key, on S3 or on the local system (default is S3)",
    )
    argument_parser.add_argument(
        "--bucket",
        required=False,
        help="If s3 is selected, the bucket name to store the key in",
    )

    return argument_parser


if __name__ == "__main__":
    parser = create_argument_parser()
    args = parser.parse_args()

    if args.datastore == "local":  # If the user chose to store the public key locally
        publisher = LocalPublisher(args.hostname, args.user)
        publisher.publish_new_key()
    else:  # If the user chose to store the public key on S3 or chose to default to S3
        if args.bucket is None:
            parser.error("The s3 option requires a bucket name!")
        publisher = S3Publisher(args.bucket, args.hostname, args.user)
        publisher.publish_new_key()



"Module for generating public/private SSH key pairs"
import os
import shutil
from Crypto.PublicKey import RSA


PRIVATE_KEY_NAME: str = "key"
PUBLIC_KEY_NAME: str = f"{PRIVATE_KEY_NAME}-cert.pub"
KEY_SIZE_BITS = 2048


def generate_private_public_key() -> tuple[bytes, bytes]:
    "Generates a private and public RSA key"
    key = RSA.generate(KEY_SIZE_BITS)
    private_key = key.export_key()
    public_key = key.public_key().export_key(format="OpenSSH")
    return private_key, public_key


def generate_private_public_key_in_file(
    public_key_dir: str,
    private_key_dir: str | None = None,
    public_key_name: str = PUBLIC_KEY_NAME,
    private_key_name: str = PRIVATE_KEY_NAME,
) -> tuple[bytes, bytes]:
    "Creates a private key and public key at the given paths"
    # If private key was not given a separate dir, use the same one as for public key
    if private_key_dir is None:
        private_key_dir = public_key_dir
    private_key, public_key = generate_private_public_key()
    user_path = os.path.expanduser("~")
    user_path_components = user_path.split("/")
    user = user_path_components[len(user_path_components) - 1]

    if not os.path.isdir(private_key_dir):
        os.makedirs(private_key_dir)
    if not os.path.isdir(public_key_dir):
        os.makedirs(public_key_dir)

    private_key_path = f"{private_key_dir}/{private_key_name}"
    public_key_path = f"{public_key_dir}/{public_key_name}"

    with open(private_key_path, "wb") as private_out:
        private_out.write(private_key)

    shutil.chown(private_key_path, user=user, group=-1)
    os.chmod(private_key_path, 0o600)

    with open(public_key_path, "wb") as public_out:
        public_out.write(public_key)
    return private_key, public_key



"SSH Server"
from typing import Callable
import os
import asyncio
import pathlib
import subprocess
from asyncio.subprocess import Process
from tempfile import NamedTemporaryFile
from key_switcheroo.server.data_stores import DataStore
from key_switcheroo.util import get_open_port
from key_switcheroo.util import get_user_path, get_username


class Server:
    "Wrapper for SSH Server. Takes a DataStore to determine where to look for keys"

    def __init__(
        self,
        data_store: DataStore,
        port: int | Callable[[], int] = get_open_port,
        authorized_key_command_executing_user: str = get_username(),
    ):
        if callable(port):
            port = port()
        self.port = port
        self.data_store = data_store
        self.authorized_key_command_executing_user = (
            authorized_key_command_executing_user
        )
        self.process: Process | None = None

    async def start(self):
        user_path = get_user_path()
        python_executable = f"{os.getcwd()}/.venv/bin/python"
        target_script = "/ssh_key_switcheroo/retrieve_public_keys.py"
        ky_cmnd = f"{python_executable} {target_script} %u {self.data_store.get_sshd_config_line()}"
        python_executable = f"{os.getcwd()}/.venv/bin/python"
        target_script = "/ssh_key_switcheroo/retrieve_public_keys.py"
        ky_cmnd = f"{python_executable} {target_script} %u {self.data_store.get_sshd_config_line()}"
        config: list[str] = [
            "LogLevel DEBUG3",
            f"Port {self.port}",
            f"HostKey {user_path}/etc/ssh/ssh_host_rsa_key",
            f"PidFile {user_path}/var/run/sshd.pid",
            "UsePAM yes",
            "AuthorizedKeysFile none",
            f"AuthorizedKeysCommand {ky_cmnd}",
            f"AuthorizedKeysCommand {ky_cmnd}",
            f"AuthorizedKeysCommandUser {self.authorized_key_command_executing_user}",
            "PasswordAuthentication no",
            "KbdInteractiveAuthentication no",
            "PubkeyAuthentication yes",
            "StrictModes yes",
        ]
        # Configuration is emitted as a temporary file to launch sshd
        with NamedTemporaryFile(mode="w+t") as temp_config:
            for option in config:
                temp_config.write(f"{option}\n")
            temp_config.file.flush()
            command: str = (
                f'/usr/sbin/sshd -f"{temp_config.name}" -E{user_path}/ssh/sshd_log'
            )
            task: Process = await asyncio.create_subprocess_shell(
                command,
                user=get_username(),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            self.process = task
            await self.process.wait()

    async def stop(self):
        "Stops the server, completely closing the process such that the port can be used"
        # If process is still running
        if not self.process is None and self.process.returncode is None:
            self.process.terminate()
            # See https://github.com/encode/httpx/issues/914
            await asyncio.sleep(1)
        kill_task = await asyncio.create_subprocess_shell(
            f"fuser -k {self.port}/tcp",
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
        await kill_task.wait()
        log_file = f"{get_user_path()}/ssh/sshd_log"
        os.remove(log_file)

    @property
    async def logs(self) -> list[str]:
        "Returns the sshd logs"
        log_file = f"{get_user_path()}/ssh/sshd_log"
        with open(log_file, mode="rt", encoding="utf-8") as logs:
            return logs.readlines()

    @staticmethod
    def __setup_host_keys():
        user_path = get_user_path()
        ssh_dir = f"{user_path}/etc/ssh"
        if not os.path.isdir(ssh_dir):
            os.mkdir(ssh_dir)
            subprocess.run("ssh-keygen -A -f ~", shell=True, check=True)

    @staticmethod
    def __setup_pid_file():
        user_path = get_user_path()
        run_dir = f"{user_path}/var/run"
        if not os.path.isdir(run_dir):
            os.mkdir(run_dir)

    def __setup_authorized_keys_script(self):
        parent_dir = pathlib.Path(__file__).parent.resolve()
        app_data_dir = "/ssh_key_switcheroo"
        python_retrieval_script_path = f"{parent_dir}/retrieve_public_keys.py"
        target_path = f"{app_data_dir}/retrieve_public_keys.py"
        if not os.path.exists(app_data_dir):
            subprocess.run(f"sudo mkdir {app_data_dir}", check=True, shell=True)
        subprocess.run(
            f"sudo cp {python_retrieval_script_path} {target_path}",
            shell=True,
            check=True,
        )
        subprocess.run(
            f"sudo chgrp 0 {python_retrieval_script_path}", check=True, shell=True
        )
        subprocess.run(
            f"sudo chmod 755 {python_retrieval_script_path}", check=True, shell=True
        )

    async def __aenter__(self):
        self.__setup_host_keys()
        self.__setup_pid_file()
        self.__setup_authorized_keys_script()
        self.__setup_authorized_keys_script()
        self.data_store.__enter__()
        await self.start()
        return self

    async def __aexit__(self, one, two, three):
        await self.stop()
        self.data_store.__exit__(None, None, None)
